import hashlib
import time
import random
import simplejson
import httplib
import re
import os
from bisect import bisect_left
import string
import ast
import functools
from random import choice
from collections import Sequence

import tornado.ioloop
import tornado.web

import requests

import redis

from db import Oliver,TryAgainSoon,DEFAULT_CONTENT_TYPE
from nginx_conf import UpstreamServer
from argparser import get_parser
from address import Address
from redis_helper import RedisSet

try:
    os.makedirs(os.path.join(os.getcwd(),'log'))
except OSError:
    pass
from log import *
LOGGER = logging.getLogger('oliver')

class Node(object):
    
    instance = None
    
    def __init__(\
                 self,
                 host,
                 port         = 80,
                 entree       = None,
                 n_replicas   = 1,
                 redisport    = 6379,
                 dbpath       = None,
                 initial_size = 1000 * 1024 * 100,
                 group_by     = lambda x : x,
                 hash_func    = lambda x : x,
                 nodename     = None,
                 location     = None,
                 secret_key   = None,
                 ip           = None,
                 public_addr  = None,
                 nginx_conf   = None):
        
        '''
        host - tornado hostname
        port - tornado port
        location - public hostname to be used for building location headers
        TODO: address - public address of this node 
        '''
        
        super(Node,self).__init__()
        self._local_address = Address(host,port)
        
        self.entree = None
        if entree:
            self.entree = Address(entree)
        
        self.public_addr = None
        if public_addr:
            self.public_addr = Address(public_addr)
        
        if nodename:
            self.nodename =  Address(nodename)
        elif self.public_addr:
            self.nodename = self.public_addr
        else:
            self.nodename = self._local_address
        
        if location:
            self.location = Address(location)
        elif self.public_addr:
            self.location = self.public_addr
        else:
            self.location = self._local_address
        
        self._id = hashlib.md5(str(self.nodename)).hexdigest()
        
        self.n_replicas = n_replicas
        self.redisport = redisport
        self.dbpath = dbpath
        self.timestamp = None
        self.initial_size = initial_size
        self.group_by = group_by
        self.hash_func = lambda x : hashlib.md5(hash_func(x)).hexdigest()
        self.bufsize = 64 * 1024
        self.nodes = dict()
        self.db = None
        self.secret_key = secret_key
        self._auth_header = hashlib.md5(self.secret_key + ip).hexdigest()
        self._nginx_conf = nginx_conf
        
        Node.instance = self
    
    @property
    def host(self):
        return self._local_address.host
    
    @property
    def port(self):
        return self._local_address.port
    
    @property
    def public_host(self):
        return \
        self.public_addr.host if self.public_addr else self._local_address.host
    
    @property
    def public_port(self):
        return \
        self.public_addr.port if self.public_addr else self._local_address.port
    
    def __enter__(self):
        return self
    
    def __exit__(self,t,value,traceback):
        if not self.db:
            return
        self.db.__exit__(t,value,traceback)
    
    def get(self,_id):
        return self.db.read_stream(_id)
    
    def put(self,_id,content_type = None):
        return self.db.write_stream(_id, content_type = content_type)
    
    def delete(self,_id,propagate = True):
        del self.db[_id]
        
        if not propagate:
            return
         
        nodes = self._nearest(_id)
        try:
            nodes.remove(self._id)
        except KeyError:
            pass
        
        for node in nodes:
            uri = self._uri(self.nodes[node],'data',_id,{'propagate' : False})
            requests.delete(uri,headers = self.internal_headers)
    
    def head(self,_id):
        return self.db.head(_id)
    
    @property
    def _ids(self):
        return self.db._ids
    
    def _uri(self,node,resource = None, _id = None, querystring = None):
        host,port = node
        uri = 'http://{host}{sep}{port}/'.format(\
            host = host,
            sep = ':' if port else '', 
            port = port or '')
        if resource:
            uri += resource + '/'
            if _id:
                uri += _id + '/'
        
        if querystring:
            uri += '?'
            uri += string.join(\
                ['{key}={value}'.format(key = k,value = v) \
                 for k,v in querystring.iteritems()],'&')
        
        return uri
    
    def put_location_response_header(self,_id):
        return self._uri(self.location,'data',_id)
    
    @property
    def _entree_address(self):
        return self._uri(self.entree)
    
    def _data_uri(self,node,_id = None):
        return self._uri(node,resource = 'data', _id = _id)
    
    def _node_uri(self,node,_id = None):
        return self._uri(node,resource = 'nodes', _id = _id)
    
    def start_server(self):
        self.timestamp = time.time()
        LOGGER.info('{nodename} coming online'.format(nodename = self.nodename))
        
        self.redis = redis.Redis(port = self.redisport)
        
        self.db = Oliver(\
                self.dbpath,
                self.redis,
                initial_size_bytes = self.initial_size,
                group_by = self.group_by)
        self.db.start()
        self._keys_to_synch = RedisSet(\
                self.redis,self.db._redis_key('keys_to_synch'))
        self.db.set_put_queue(self._keys_to_synch)
        
        self._join_network()
        
        self._healthcheck_job = \
            tornado.ioloop.PeriodicCallback(self._healthcheck,5000)
        self._healthcheck_job.start()
        
        self._rebalance_job = \
            tornado.ioloop.PeriodicCallback(self._rebalance,2500)
        self._rebalance_job.start()
    
    @property
    def json_serializable_nodes(self):
        return dict(((k,tuple(v)) for k,v in self.nodes.iteritems()))
    
    @property
    def _handshake_message(self):
        return simplejson.dumps({
                'host' : self.public_host, 
                'port' : self.public_port,
                'entree' : str(self.entree) if self.entree else None,
                'n_replicas' : self.n_replicas,
                'timestamp' : self.timestamp,
                'nodes' : self.json_serializable_nodes,
                'pid' : os.getpid(),
                'nodename' : self._id})
    
    @property
    def internal_headers(self):
        return {'Content-Type' : 'application/json',
                'X-Oliver'     : self._auth_header,
                'X-Oliver-Balanced' : self.nodename}
    
    def _send_handshake(self,node_addr):
        r = requests.put(self._node_uri(node_addr, _id = self._id),
                        data = self._handshake_message,
                        headers = self.internal_headers)
        return r
    
    def _send_node_down(self,down_id, recipient_node):
        r = requests.delete(\
                self._node_uri(recipient_node,_id = down_id),
                headers = self.internal_headers)
        return r
    
    def _join_network(self):
        if not self.entree:
            return
        
        # get the entree node's id
        r = requests.get(self._uri(self.entree),headers = self.internal_headers)
        if not r.ok:
            raise Exception('entree responded with status code {status_code}'\
                            .format(status_code = r.status_code))
        entree_config = r.json()
        entree_id = entree_config['nodename']
        self.nodes[entree_id] = self.entree
        
        # get the full list of nodes that my entree knows about
        r = requests.get(self._node_uri(self.entree),headers = self.internal_headers)
        if not r.ok:
            raise Exception('entree responded with status code {status_code}'\
                            .format(status_code = r.status_code))
            
        # update my node list with all the nodes that my entree knows about
        self.nodes.update(r.json())
        self._reload_nginx()
        
        # add myself to my entree's node list
        self._send_handshake(self.entree)
        
        # iterate over all the new nodes and introduce myself
        for node_id,address in self.nodes.iteritems():
            if node_id == entree_id:
                continue
            self._send_handshake(address)
    
    def _add_node(self,config):
        host,port = config['host'],config['port']
        _id = config['nodename']
        self.nodes[_id] = Address(host,port)
        if config['timestamp'] > self.timestamp:
            self.n_replicas = config['n_replicas']
        self._reload_nginx()
        ring = self._hash_ring()
        _ids = self._ids
        if ring.affected(_id, self.n_replicas) and _ids:
            self._keys_to_synch.add(*_ids)
        return _id
    
    def _delete_node(self,_id):
        del self.nodes[_id]
        ring = self._hash_ring(_id)
        _ids = self._ids    
        if ring.affected(_id, self.n_replicas) and _ids:
            self._keys_to_synch.add(*_ids)
        self._reload_nginx()
    
    # TODO: I shouldn't know about nginx here. This should be a callback that 
    # gets passed in
    def _reload_nginx(self):
        if not self._nginx_conf:
            return
        
        servers = [UpstreamServer(self.host,self.port)]
        servers.extend([UpstreamServer(*Address(*n)) \
                        for n in self.nodes.itervalues()])
        
        self._nginx_conf.main_upstream.n_replicas = self.n_replicas
        self._nginx_conf.main_upstream.tornado_upstreams = servers
        self._nginx_conf.reload_conf()
        
    def _healthcheck(self):
        if not len(self.nodes):
            return
        
        _id = random.choice(self.nodes.keys())
        node = self.nodes[_id] 
        uri = self._node_uri(node)
        r = None
        try:
            r = requests.get(uri,headers = self.internal_headers)
        except requests.ConnectionError:
            pass
        
        if not r or not r.status_code in (httplib.OK,httplib.SERVICE_UNAVAILABLE):
            del self.nodes[_id]
            for address in self.nodes.itervalues():
                self._send_node_down(_id, address)
    
    def _ensure_node_has_data(self,meta,node_id,_id):
        LOGGER.info('IN ENSURE NODE HAS DATA')
        addr = self.nodes[node_id]
        uri = self._data_uri(addr, _id)
        LOGGER.info('MAKING HEAD REQUEST FOR %s from %s' % (_id,addr))
        r = requests.head(uri,headers = {'X-Oliver-Balanced' : self.nodename})
        LOGGER.info('MADE HEAD REQUEST FOR %s from %s' % (_id,addr))
        
        if r.status_code == httplib.ACCEPTED:
            return
        
        # TODO: Get rid of "zounds" here
        if r.status_code == httplib.NO_CONTENT and \
            (r.headers['etag'] == meta.hexdigest or \
             meta.timestamp <= int(r.headers['x-zounds-usecs'])):
            # the HEAD request was successful, and either the etags
            # matched or the other node has a more recent version than
            # we do, so, do nothing.
            return
        try:
            LOGGER.info('self : {s}, other : {o}'.format(\
                                s = meta.hexdigest,o = r.headers['etag']))
            LOGGER.info('self : {s}, other : {o}'.format(\
                                s = meta.timestamp,o = r.headers['x-zounds-usecs']))
        except KeyError:
            pass
        
        LOGGER.info('STARTING TO PUT {id} to {node}'.format(id = _id,node = addr))
        
        # either the other node did not have a copy, or its copy is old.
        with self.db.read_stream(_id) as rs:
            requests.put(uri,
                         data = chunked(rs,bufsize = self.bufsize),
                         headers = {'Content-Type' : meta.content_type,
                                    'X-Oliver'     : self._auth_header,
                                    'X-Oliver-Balanced' : self.nodename})
    
    def _get_next_key_to_synch(self):
        key = self._keys_to_synch.pop()
        if key:
            return key    
        try:
            return choice(self.db._ids)
        except IndexError:
            return None
    
    def _hash_ring(self,*additional):
        nodes = self.nodes.keys()
        nodes.append(self._id)
        nodes.extend(additional)
        return Ring(nodes)
    
    def _nearest(self,_id):
        ring = self._hash_ring()
        start = ring.bisect(self.hash_func(_id))    
        return set(ring[start : start + self.n_replicas])
    
    def _rebalance(self):
        LOGGER.info('IN REBALANCE')
        # TODO: The number of keys I process here should be proportional to the
        # number of items in the queue
        _id = self._get_next_key_to_synch()
        if not _id:
            LOGGER.info('_id was None')
            return
        
        LOGGER.info('REBALANCING %s' % _id)
        
        nodes = self._nearest(_id)
        delete = False
        try:
            nodes.remove(self._id)
        except KeyError:
            delete = True
        
        meta = None
        try:
            meta = self.db.head(_id)
        except (KeyError,TryAgainSoon):
            return
        
        for node in nodes: 
            self._ensure_node_has_data(meta, node, _id)
            LOGGER.info('pushed {_id} to {node}'.format(**locals()))
                            
        if delete:
            del self.db[_id]
            LOGGER.info('deleting {_id} locally'.format(**locals()))
        
        LOGGER.info('DONE REBALANCING %s' % _id)
 
def oliver_auth(f):
    
    @functools.wraps(f)
    def x(inst,*args,**kwargs):
        # KLUDGE: This is to help diagnose some other problems I'm having, and
        # needs to go away
        if 'oliver' == Node.instance.secret_key:
            f(inst,*args,**kwargs)
            return
        
        try:
            ip = inst.request.headers['X-Real-IP']
        except KeyError:
            ip = inst.request.remote_ip
        
        if ip in ['::1','127.0.0.1']:
            ip = '127.0.0.1'
        
        try:
            secret = inst.request.headers['X-Oliver']
            if secret != hashlib.md5(Node.instance.secret_key + ip).hexdigest():
                inst.set_status(httplib.FORBIDDEN)
                inst.finish()
                return
            f(inst,*args,**kwargs)
        except KeyError:
            inst.set_status(httplib.FORBIDDEN)
            inst.finish()
    
    return x

@tornado.web.stream_body
class DataHandler(tornado.web.RequestHandler):
    
    def options(self):
        raise NotImplemented()
    
    def _set_headers(self,meta,**kwargs):
        # TODO: Get rid of "zounds" here
        self.set_header('X-Zounds-Usecs',meta.timestamp)
        self.set_header('Content-Type',meta.content_type)
        self.set_header('Accept-Ranges','bytes')
        for k,v in kwargs.iteritems():
            self.set_header(k.replace('_','-'),v)
    
    def head(self,_id):
        meta = None
        
        LOGGER.info('HTTP HEAD request for %s' % _id)
        try:
            meta = Node.instance.head(_id)
            LOGGER.info('HTTP HEAD found %s' % _id)
        except KeyError:
            LOGGER.info('HTTP HEAD 404 %s' % _id)
            self.set_status(httplib.NOT_FOUND)
            self.finish()
            return
        except TryAgainSoon:
            LOGGER.info('HTTP HEAD Try Again Soon %s' % _id)
            self.set_status(httplib.ACCEPTED)
            self.finish()
            return
        
        self.meta = meta
        self._set_headers(meta,ETag = meta.hexdigest)
        self.set_status(httplib.NO_CONTENT)
        self.finish()
    
    def get(self,_id = None):
        if not _id:
            self.set_header('Content-Type','application/json')
            self.write(simplejson.dumps({'keys' : Node.instance._ids}))
            self.finish()
            return
        
        try:    
            with Node.instance.get(_id) as stream:
                meta = stream.meta
                rr = RangeRequest(self.request.headers,meta)
                cr = rr.content_range
                self.meta = meta
                self._set_headers(meta,Content_Length = rr.content_length)
                self.set_status(rr.response_code)
                if cr:
                    self.set_header('Content-Range',cr)
                
                stream.seek(rr.start)
                self.write(stream.read(rr.content_length))
            self.finish()
            return
        except KeyError:
            self.send_error(httplib.NOT_FOUND)
            return
        except TryAgainSoon:
            self.send_error(httplib.ACCEPTED)
            return
        
    
    def compute_etag(self):
        try:
            return self.meta.hexdigest
        except:
            pass
    
    @oliver_auth
    def put(self,_id):
        LOGGER.info('HTTP Start PUT %s' % _id)
        content_type = DEFAULT_CONTENT_TYPE
        try:
            content_type = self.request.headers['Content-Type']
        except KeyError:
            pass
        
        self._f = Node.instance.put(_id,content_type = content_type)
        self.data_key = _id
        
        try:
            content_length = int(self.request.headers['Content-Length'])
            LOGGER.info('HAD CONTENT LENGTH')
            self.request.connection.stream.read_bytes(\
                                            content_length,self._got_entirety)
        except KeyError:
            LOGGER.info('CHUNKED CONTENT')
            self.request.request_continue()
            self.request.connection.stream.read_until(\
                                        '\r\n',self._got_size)
    
    @oliver_auth
    @tornado.web.asynchronous
    def delete(self,_id):
        p = ast.literal_eval(self.get_argument('propagate', str(True)))
        tornado.ioloop.IOLoop.instance().add_callback(self._delete,_id,p)
        self.set_status(httplib.NO_CONTENT)
        self.finish()
    
    def _delete(self,_id,p):
        Node.instance.delete(_id,p)
    
    @property
    def _location(self):
        return Node.instance.put_location_response_header(self.data_key)
    
    def _got_entirety(self,chunk = ''):
        LOGGER.info('GOT ENTIRETY')
        self._f.write(chunk)
        self._f.close()
        self.set_status(httplib.CREATED)
        self.set_header('Location',self._location)
        self.finish()
        return
    
    def _got_size(self,chunk = ''):
        LOGGER.info('GOT SIZE')
        n_bytes = int(chunk[:-2],16)
        if not n_bytes:
            self._f.flush()
            self._f.close()
            self.set_status(httplib.CREATED)
            self.set_header('Location',self._location)
            self.finish()
            return
        
        self.request.connection.stream.read_bytes(\
                                    n_bytes + 2,self._got_chunk)
    
    def _got_chunk(self,chunk = ''):
        LOGGER.info('GOT CHUNK')
        self._f.write(chunk[:-2])
        self._f.flush()
        self.request.connection.stream.read_until(\
                                    '\r\n',self._got_size)


class NodeHandler(tornado.web.RequestHandler):
    
    @oliver_auth
    def options(self):
        raise NotImplemented()
    
    @oliver_auth
    def put(self,_id):
        '''
        Add a new node to my nodelist.
        '''
        config = simplejson.loads(self.request.body)
        Node.instance._add_node(config)
        self.set_status(httplib.CREATED)
        self.set_header('Location',self.request.uri)
        self.finish()
        
    # TODO: Don't set E-Tags, and attempt to get clients to not cache responses
    # TODO: Show a simple list of nodes to unauthenticated clients when id is 
    # not specified
    @oliver_auth        
    def get(self,_id = None):
        '''
        If _id is specified, return the configuration for the specified node.
        Otherwise, list nodes.
        '''
        # TODO: factor out the content type here
        self.set_header('Content-Type','application/json')
        if not _id:
            self.write(simplejson.dumps(Node.instance.json_serializable_nodes))
            self.finish()
            return
        
        # TODO: Does this actually work?  I don't think this is valid JSON
        try:
            self.write(simplejson.dumps(Node.instance.nodes[_id]))
        except KeyError:
            self.set_status(httplib.NOT_FOUND)
        self.finish()
    
    @oliver_auth
    def delete(self,_id):
        '''
        Remove the node from my node collection
        '''
        Node.instance._delete_node(_id)
        self.set_status(httplib.NO_CONTENT)
        self.finish()

class MainHandler(tornado.web.RequestHandler):
    
    @oliver_auth
    def get(self):
        self.set_header('Content-Type','application/json')
        self.write(Node.instance._handshake_message)
        self.finish()

class DebugHandler(tornado.web.RequestHandler):
    def get(self):
        self.set_header('Content-Type','application/json')
        data = dict()
        data['headers'] = dict(self.request.headers)
        body = simplejson.dumps(data)
        self.write(body)
        self.finish()
        

# TODO: tornado static file module contains code to parse range requests, I think
class RangeRequest(object):
    
    RANGE = 'Range'
    RE = re.compile(r'bytes\s*=\s*(?P<start>\d+)\s*-\s*(?P<stop>\d+)?')
    
    def __init__(self,headers,metadata):
        super(RangeRequest,self).__init__()
        self.headers = headers
        self.metadata = metadata
        self.response_code = None
        self._content_length = None
        self.start = None
        self.stop = None
        self._parse()
    
    def _parse(self):
        if (self.RANGE not in self.headers) or not self.headers[self.RANGE]:
            self.start = 0
            self.stop = self.metadata.size - 1
            self.response_code = httplib.OK
            self.content_length = self.metadata.size
            return
        
        m = self.RE.match(self.headers[self.RANGE])
        if not m:
            self.response_code = httplib.REQUESTED_RANGE_NOT_SATISFIABLE
            return
        d = m.groupdict()
        self.start = int(d['start'])
        if self.start >= self.metadata.size:
            self.response_code = httplib.REQUESTED_RANGE_NOT_SATISFIABLE
            return
        try:
            self.stop = int(d['stop'])
        except TypeError:
            self.stop = self.metadata.size - 1
        self.content_length = (self.stop - self.start) + 1
        self.response_code = httplib.PARTIAL_CONTENT
    
    @property
    def content_range(self):
        if httplib.PARTIAL_CONTENT != self.response_code:
            return None
        
        return 'bytes {start}-{stop}/{total}'.format( \
                        start = self.start,
                        stop = self.stop,
                        total = self.metadata.size)
        
def chunked(f,bufsize = 1000 * 64):
    data = f.read(bufsize)
    if data: yield data


class Ring(Sequence):
    
    def __init__(self,l):
        l.sort()
        self.data = l
    
    def __len__(self):
        return len(self.data)
    
    def index(self,item):
        return self.data.index(item)
    
    def bisect(self,item):
        return bisect_left(self.data,item)
    
    def affected(self,_id,n_replicas):
        index = self.index(_id)
        r = n_replicas - 1
        start = index - min(1,r)
        stop = index + r
        return _id in self[start : index] or _id in self[index : stop]
    
    def __getitem__(self,index):
        if isinstance(index,int):
            return self.data[index % len(self)]
        
        if isinstance(index,slice):
            if index.step:
                raise NotImplemented()
            index = slice(index.start or 0,index.stop or len(self))
            x = max(0,index.stop - index.start)
            if not x: return []
            return [self[(index.start % len(self)) + i] for i in xrange(x)]
            
        raise NotImplemented()

class OliverApplication(tornado.web.Application):
    
    def __init__(self,*args,**kwargs):
        super(OliverApplication,self).__init__(*args,**kwargs)
    
    def log_request(self,req):
        try:
            remote_ip = req.request.headers['X-Real-IP']
        except KeyError:
            remote_ip = req.request.remote_ip
        
        msg = '{method} {uri} from {remote_ip}'.format(\
            method = req.request.method,
            uri = req.request.uri,
            remote_ip = remote_ip)
        LOGGER.info(msg)

application = OliverApplication([
    (r'/',MainHandler),
    (r'/nodes/?([^/]+)?/?',NodeHandler),
    (r'/data/?([^/]+)?/?',DataHandler),
    (r'/debug/?',DebugHandler)
], debug = True)

if __name__ == '__main__':
    parser = get_parser()
    args = parser.parse_args()
    with Node(\
        args.host,
        args.port,
        args.entree,
        args.replicas,
        args.redisport,
        args.dbpath,
        args.initialsize,
        eval(args.groupby),
        eval(args.hash),
        args.nodename,
        args.location,
        args.secret_key,
        args.ip,
        args.public_addr) as node:
        node.start_server()
        application.listen(node.port)
        tornado.ioloop.IOLoop.instance().start()
    